"""
Metrics for the trimkmeans algorithm
These are wrapper functions that remove the trimmed cluster marked by the highest label integer
"""

from numpy import var
from sklearn import metrics


def sum_variances(data, labels, cluster):
    """
    :param data: list of datapoints with trimmed points removed
    :param labels: list of labels generated by trimkmeans.predict() with trimmed points removed
    :param cluster: amount of clusters in the dataset
    :return: sum of variances in each cluster
    """
    # combine data and label so the label can be matched to the point
    zipped = zip(data, labels)
    zipped_list = list(zipped)
    sum_var = 0
    for i in range(cluster):
        sum_var += var([x[0] for x in zipped_list if x[1] == i])
    return sum_var


def score_even_distribution(data, labels, cluster):
    """
    :param data: list of datapoints with trimmed points removed
    :param labels: list of labels generated by trimkmeans.predict() with trimmed points removed
    :param cluster: amount of clusters in the dataset
    :return: product of count of points in each cluster
    """
    # combine data and label so the label can be matched to the point
    zipped = zip(data, labels)
    zipped_list = list(zipped)
    prod_cnt = 0
    for i in range(cluster):
        prod_cnt *= len([x[0] for x in zipped_list if x[1] == i])
    return prod_cnt


def trimmed_kmeans_metric_unsupervised(x_train, labels, metric):
    """
    remove all the trimmed points (label = n_centers) since they would interfere with the Silhouette score
    :param x_train: list of datapoints
    :param labels: list of labels generated by trimkmeans.predict()
    :param metric: name of the metric which should be evaluated from ['silhouette_score', 'sed', 'sv']
    :return:
    """
    # combine data and label so they can be deleted together with the order still in tact
    zipped = zip(x_train, labels)
    zipped_list = list(zipped)
    # the highest label is for the trimmed points
    centers = max(labels)
    removed_trimmed = [i for i in zipped_list if i[1] != centers]
    x_train_removed_trimmed, classification_removed_trimmed = list(zip(*removed_trimmed))
    if metric == 'silhouette_score':
        return metrics.silhouette_score(x_train_removed_trimmed, classification_removed_trimmed, metric='euclidean')
    if metric == 'sed':
        return score_even_distribution(x_train_removed_trimmed, classification_removed_trimmed, centers)
    if metric == 'sv':
        return sum_variances(x_train_removed_trimmed, classification_removed_trimmed, centers)

    raise ValueError("metric must be either 'silhouette_score', 'sed' or 'sv")


#
def trimmed_kmeans_metric_supervised(true_labels, labels, metric):
    """
        remove all the trimmed points labels (label = n_centers) since they would interfier with the supervised metric
        :param true_labels: list of true labels of a dataset, for example generated by sklearn.dataset.make_blobs()
        :param labels: list of labels generated by trimkmeans.predict()
        :param metric: name of the metric which should be evaluated from ['rand_score','completeness_score']
        :return:
        """
    # combine data and label so they can be deleted together with the order still in tact
    zipped = zip(true_labels, labels)
    zipped_list = list(zipped)
    # the highest label is for the trimmed points
    centers = max(labels)
    removed_trimmed = [i for i in zipped_list if i[1] != centers]
    true_label_removed_trimmed, classification_removed_trimmed = list(zip(*removed_trimmed))
    if metric == 'rand_score':
        return metrics.rand_score(true_label_removed_trimmed, classification_removed_trimmed)
    if metric == 'completeness_score':
        return metrics.completeness_score(true_label_removed_trimmed, classification_removed_trimmed)
    raise ValueError("metric must be either 'rand_score' or 'completeness_score'")



"""
unittests for packaging
"""
import heapq
import unittest

import numpy as np

from src.trimkmeans.metrics import trimmed_kmeans_metric_supervised
from src.trimkmeans.metrics import trimmed_kmeans_metric_unsupervised
from src.trimkmeans.trimkmeans import TrimKMeans


class TestingTrimKMeans(unittest.TestCase):
    """
    tests for the TrimKMeans class
    """

    def test_private_create_points(self):
        """
        asserts if ClusterPoints created by TrimKMeans private method are created like expected
        :return: None
        """
        trimkmeans = TrimKMeans()
        x_train = np.array([[1, 1], [2, 2], [3, 3]])
        centroids = np.array([[1, 1], [2, 2], [3, 3]])
        cp1 = trimkmeans.ClusterPoint(np.array([1, 1]))
        cp2 = trimkmeans.ClusterPoint(np.array([2, 2]))
        cp3 = trimkmeans.ClusterPoint(np.array([3, 3]))
        cp1.dist, cp2.dist, cp3.dist = -0.0, -0.0, -0.0
        cp1.cluster, cp2.cluster, cp3.cluster = 0, 1, 2
        sorted_points = []
        heapq.heappush(sorted_points, cp1)
        heapq.heappush(sorted_points, cp2)
        heapq.heappush(sorted_points, cp3)
        test_data = trimkmeans._TrimKMeans__create_points(x_train, centroids)
        self.assertIsNone(np.testing.assert_array_equal(np.array(sorted_points), np.array(test_data)))

    def test_empty_data(self):
        """
        Tests if empty input data raises the right error
        :return: None
        """
        trimkmeans = TrimKMeans(n_clusters=3)
        testdata = np.array([])
        with self.assertRaises(ValueError):
            trimkmeans.fit(testdata)

    def test_cluster_one_dim(self):
        """
        Tests if fit() method works as expected if points are one-dimensional
        :return: None
        """
        trimkmeans = TrimKMeans(n_clusters=3, trim=0.3, n_init=1000)
        testdata = np.array([[-100.], [0.], [1.], [100.]])
        try:
            trimkmeans.fit(testdata)
        except RuntimeError:
            self.fail("fit() with one dimensional data failed")

    def test_init_with_centroids(self):
        """
        Tests if fit() method works as expected if points are provided via the init parameter
        :return: None
        """
        trimkmeans = TrimKMeans(n_clusters=3, trim=0.3, n_init=1, init=np.array([[0., 1.], [1., 2.], [2., 3.]]))
        testdata = np.array([[0., 1.], [1., 2.], [2., 3.], [3., 4.]])
        try:
            trimkmeans.fit(testdata)
            trimkmeans.predict(testdata)
        except RuntimeError:
            self.fail("fit() with init centroids failed")

    def test_less_points_then_cluster(self):
        """
        Tests if a size error is raised which stems from a dataset with fewer points than clusters specified
        :return: None
        """
        trimkmeans = TrimKMeans(n_clusters=4)
        testdata = np.array([[1], [2], [3]])
        with self.assertRaises(ValueError):
            trimkmeans.fit(testdata)

    def test_trim_to_big(self):
        """
        Tests if a size error is raised even if the size error is created during trimming
        :return: None
        """
        trimkmeans = TrimKMeans(n_clusters=3, trim=0.4)
        testdata = np.array([[1], [2], [3]])
        with self.assertRaises(ValueError):
            trimkmeans.fit(testdata)


class TestingMetrics(unittest.TestCase):
    """
    Tests for the trimmed kmeans metrics from trimkmeans.metrics
    """

    def test_trimmed_kmeans_silhouette(self):
        """
        Tests the silhouette score metric for trimmed data
        :return:
        """
        self.assertEqual(1, trimmed_kmeans_metric_unsupervised([[0], [1], [2], [1.5], [2], [1], [0]],
                                                               [0, 1, 2, 3, 2, 1, 0],
                                                               'silhouette_score'))

    def test_trimmed_kmeans_rand(self):
        """
        Tests the rand score metric for trimmed data
        :return:
        """
        self.assertEqual(1, trimmed_kmeans_metric_supervised([0, 1, 2], [0, 1, 2], 'rand_score'))


if __name__ == '__main__':
    unittest.main()



"""
the example in R is a single function which takes an object of class 'tkm' as a parameter
to be more in line with the KMeans implementation in Pythons SKlearn library the following implementation
is realized as a class
"""
import random
import warnings
from math import inf, floor

import numpy as np
from sklearn.preprocessing import StandardScaler


def euclidean(point, data):
    """
    Euclidean distance between point & data.
    Point has dimensions (m,), data has dimensions (n,m), and output will be of size (n,).
    """
    # list conversion to prevent deprecation warning Calling np.sum(generator) is deprecated
    gen = ((point - data) ** 2)
    return np.sqrt(np.sum(gen, axis=1))


class TrimKMeans:
    """
    Optimizes the k-means algorithm by trimming a portion of the points.
    The implementation follows the algorithm of the r package but is modeled after the scikit-learn
    kmeans clustering code.
    The centroids are initialized using the "k-means++" method, where a random datapoint is selected as the first,
    then the rest are initialized with probabilities proportional to their distances to the first
    """

    # replaced parameters countmode and printcrit with sklearns verbose level
    # replaced R's run and maxit Parameters with sklearns n_init and max_iter
    def __init__(self,
                 n_clusters=8,
                 trim=0.1,
                 scaling=False,
                 n_init=10,
                 max_iter=100,
                 verbose=0,
                 random_state=None,
                 init='k-means++'):
        self.n_clusters = n_clusters
        self.trim = trim
        self.scaling = scaling
        self.n_init = n_init
        self.max_iter = max_iter
        self.verbose = verbose
        self.random_state = random_state
        self.opt_cutoff_ranges = None
        self.cluster_centers_ = None
        self.crit_val = -1 * inf
        self.init = init

    class ClusterPoint:
        """
        Inner Class used to group points together with the distance and cluster that was calculated by the algorithm
        """
        # https://stackoverflow.com/questions/53388451/how-to-speed-up-python-instance-initialization-for-millions-of-objects
        # 107.801 -> 92.025 cumtime for 10000 points
        __slots__ = ('points', 'cluster', 'dist')

        def __init__(self, points=None):
            self.points = points
            self.cluster = None
            self.dist = None

        # enable comparison for sort
        def __lt__(self, cp2):
            return self.dist < cp2.dist

        # for debugging
        def __repr__(self):
            return f"Point: {self.points}, Cluster: {self.cluster}, Distance: {self.dist}"

        # for testing
        def __eq__(self, cp2):
            if self.cluster == cp2.cluster and self.dist == cp2.dist:
                return np.array_equal(self.points, cp2.points)
            return False

    def __create_points(self, x_train, centroids):
        """
        Inner function used to group the transformation of raw point data to instances of inner class ClusterPoint
        :param x_train: list of datapoints
        :param centroids: centers of kmeans cluster
        :return: heapq of sorted ClusterPoint instances
        """
        sorted_points = np.array([self.ClusterPoint() for _ in range(len(x_train))])
        # Sort each datapoint, assigning to nearest centroid
        for idx, val in enumerate(x_train):
            sorted_points[idx].points = val
            # save the distance for each point for trimming later
            # distance is negated because heapq is implemented as a min stack
            dists = -1 * euclidean(val, centroids)
            sorted_points[idx].dist = max(dists)
            sorted_points[idx].cluster = np.argmax(dists)

        # trim the n points
        sorted_points.sort()
        for x in range(0, floor(self.trim * len(sorted_points))):
            sorted_points[x].cluster = len(centroids)

        return sorted_points

    def __calculate_cutoff(self, sorted_points):
        """
        # set the member variable cutoff range which is the last point in a cluster not cut off
        :param sorted_points: result of the __create_points() method
        :return:
        """
        self.opt_cutoff_ranges = [None] * self.n_clusters

        for i in range(self.n_clusters):
            # get max dist per cluster
            try:
                self.opt_cutoff_ranges[i] = min(x.dist for x in sorted_points if x.cluster == i)
            except ValueError:
                self.opt_cutoff_ranges[i] = -1 * inf

    def __compare_runs(self, sorted_points, centroids, run):
        """
        Encapsulates the comparison between the different results
         of two runs of trimmed_kmeans
        :param sorted_points: np.array of ClusterPoints
        :param centroids: centroids of the clusters
        :param run: number of iteration for verbose printing
        """
        # calculate the sum of all the distances
        new_crit_val = sum((c_p.dist for c_p in sorted_points))
        if self.verbose >= 1:
            print(f"Run {run} criterion value {new_crit_val}")
        if new_crit_val > self.crit_val:
            self.__calculate_cutoff(sorted_points)
            self.crit_val = new_crit_val
            self.cluster_centers_ = centroids

    def __check_data(self, data):
        """
        throws ValueError if the input data has the wrong shape
        :param self:
        :param data: list of data points to fit the trimmed kmeans to
        :return:
        """
        if not isinstance(data, np.ndarray):
            raise ValueError(
                "Please convert input data to numpy array"
            )
        # If input is scalar raise error
        if len(data) == 0:
            raise ValueError(
                "Expected 2D array, array is empty"
            )
        # If input is 1D raise error
        if len(data[0]) == 0:
            raise ValueError(
                "Expected 2D array, got 1D array instead"
            )
        # If input n_rows is smaller than n_cluster raise Error
        trimmed_len = len(data) - floor(self.trim * len(data))
        if trimmed_len < self.n_clusters:
            raise ValueError(
                f"n_samples-trim={trimmed_len} should be >= n_clusters={self.n_clusters}."
            )

    def fit(self, x_train):
        """computes trimmed k means clustering
        https://towardsdatascience.com/create-your-own-k-means-clustering-algorithm-in-python-d7d4c9077670
        :param x_train: list of datapoints
        """
        self.__check_data(x_train)
        if self.scaling:
            x_train = StandardScaler().fit_transform(x_train)
        if self.random_state:
            random.seed(self.random_state)

        for run in range(self.n_init):
            if isinstance(self.init, str) and self.init == "k-means++":
                # Pick a random point from train data for first centroid
                centroids = [random.choice(x_train)]
                for _ in range(self.n_clusters - 1):
                    # Calculate distances from points to the centroids
                    dists = sum((euclidean(centroid, x_train) for centroid in centroids))
                    # Normalize the distances
                    dists /= np.sum(dists)
                    # Choose remaining points based on their distances
                    new_centroid_idx, = np.random.choice(range(len(x_train)), size=1, p=dists)
                    centroids += [x_train[new_centroid_idx]]
            elif isinstance(self.init, str) and self.init == "random":
                # Randomly select centroid start points, uniformly distributed across the domain of the dataset
                min_point, max_point = np.min(x_train, axis=0), np.max(x_train, axis=0)
                centroids = [np.random.uniform(min_point, max_point) for _ in range(self.n_clusters)]
            elif isinstance(self.init, str):
                raise ValueError(f"{self.init} is not a recognized method, try 'k-means++', 'random' or passing"
                                 f"an inital centroid ndarray")
            else:
                # calculation for set centroids, only one iteration
                if self.n_init > 1:
                    warnings.warn("If initial mean vectors are specified, only one run will be calculated")
                centroids = self.init.copy()

            # Iterate, adjusting centroids until converged or until passed max_iter
            iteration = 0
            prev_centroids = None
            sorted_points = []
            while np.not_equal(centroids, prev_centroids).any() and iteration < self.max_iter:
                sorted_points = self.__create_points(x_train, centroids)
                # Push current centroids to previous, reassign centroids as mean of the points belonging to them
                # copy list by value[:]
                prev_centroids = centroids.copy()
                for i in range(self.n_clusters):
                    with warnings.catch_warnings():
                        # clusters can be empty, the corresponding RuntimeWarning is suppressed here
                        # to prevent screen clutter
                        warnings.filterwarnings('ignore', category=RuntimeWarning)
                        centroids[i] = np.mean([c_p.points for c_p in sorted_points if c_p.cluster == i], axis=0)
                for i, centroid in enumerate(centroids):
                    if np.isnan(centroid).any():  # Catch any np.nans, resulting from a centroid having no points
                        centroids[i] = prev_centroids[i]
                iteration += 1
            if self.verbose >= 1:
                print(f"itertions in run {run}: {iteration}")

            self.__compare_runs(sorted_points, centroids, run)

    def predict(self, X):
        """
        :param X: list of datapoints to be clustered
        :returns: cluster centroids, cluster label for the datapoints and cutoff ranges for each cluster
        """
        centroid_idxs = []
        for x in X:
            dists = euclidean(x, self.cluster_centers_)
            centroid_idx = np.argmin(dists)
            # check if distance is smaller than cutoff of that cluster
            # if not, label n_clusters is given
            if self.opt_cutoff_ranges[centroid_idx] * -1 < dists[centroid_idx]:
                centroid_idxs.append(self.n_clusters)
            else:
                centroid_idxs.append(centroid_idx)
        if self.verbose >= 1:
            print(f"trimmed k-means: trim= {self.trim} , n_clusters= {self.n_clusters}")
            print(f"Classification (trimmed points are indicated by  {self.n_clusters} ):")
        return centroid_idxs
